//========================================================================
// GLFW - An OpenGL framework
// File:     init.c
// Platform: X11 (Unix)
// Version:  2.3
// Date:     2002.11.23
// Author:   Marcus Geelnard (marcus.geelnard@home.se)
// WWW:      http://hem.passagen.se/opengl/glfw/
//------------------------------------------------------------------------
// Copyright (c) 2002 Marcus Geelnard
//
// GLFW is provided "as-is", without any express or implied warranty,
// without even the implied warranty of fitness for a particular purpose.
// In no event will the author be held liable for any damage araising from
// the use of GLFW.
//
// Permission is granted to anyone to use GLFW for any purpose, including
// commercial applications, and to alter it and redistribute it freely,
// subject to the following restrictions:
//
// 1. You must not claim that you wrote GLFW.
// 2. If you modify any parts of GLFW, it must be clearly stated which
//    modifications have been made, and that those modifications are not
//    the work of the author of GLFW.
// 3. If you use GLFW in a product, an acknowledgment in the product
//    documentation would be appreciated, but is not required.
// 4. This notice may not be removed or altered from any source
//    distribution of GLFW.
//
// Marcus Geelnard
// marcus.geelnard@home.se
//========================================================================

#define  _init_c_
#include "internal.h"



//************************************************************************
//****                  GLFW internal functions                       ****
//************************************************************************

//************************************************************************
//****                    GLFW user functions                         ****
//************************************************************************

//========================================================================
// glfwInit() - Initialize various GLFW state.
//========================================================================

int glfwInit( void )
{
    // Is GLFW already running?
    if( _glfwInitialised )
    {
        return GL_TRUE;
    }
    
    _glfwInitDyn();

    // Flag that window is not yet opened
    _glfwWin.Opened = GL_FALSE;

    // Inactivate OpenGL extension GLX_SGI_swap_control (activated by
    // glfwOpenWindow)
    _glXSwapIntervalSGI = (GLXSWAPINTERVALSGI_T) 0;

    // Clear window hints
    _glfwWinHints.RefreshRate    = 0;
    _glfwWinHints.AccumRedBits   = 0;
    _glfwWinHints.AccumGreenBits = 0;
    _glfwWinHints.AccumBlueBits  = 0;
    _glfwWinHints.AccumAlphaBits = 0;
    _glfwWinHints.AuxBuffers     = 0;
    _glfwWinHints.Stereo         = 0;

    // Install atexit() routine
    atexit( glfwTerminate );

    // Initialize joysticks
    _glfwInitJoysticks();

    // Form now on, GLFW state is valid
    _glfwInitialised = GL_TRUE;

    return GL_TRUE;
}


//========================================================================
// glfwTerminate() - Close window
//========================================================================

void glfwTerminate( void )
{
    // Is GLFW initialised?
    if( !_glfwInitialised )
    {
        return;
    }

    // Close OpenGL window
    glfwCloseWindow();

    // Initialize joysticks
    _glfwTerminateJoysticks();

    // GLFW is no longer initialized
    _glfwInitialised = GL_FALSE;
}


//========================================================================
// glfwGetVersion() - Get GLFW version
//========================================================================

void glfwGetVersion( int *major, int *minor, int *rev )
{
    if( major != NULL ) *major = GLFW_VERSION_MAJOR;
    if( minor != NULL ) *minor = GLFW_VERSION_MINOR;
    if( rev   != NULL ) *rev   = GLFW_VERSION_REVISION;
}
