#include <SDL/SDL.h>
#include <caml/mlvalues.h>
#include <caml/alloc.h>
#include <caml/memory.h>

#include "shared.h"

value SDL_event_to_ML_tevent(SDL_Event event)
{
    CAMLparam0 ();
    CAMLlocal2(ML_event, to_return);
    int i;

    switch (event.type) 
    {
	case SDL_ACTIVEEVENT: 
	{
    	    ML_event=alloc(2, 0);
	    Store_field(ML_event, 0, Val_int(event.active.gain));
	    if (event.active.state & SDL_APPACTIVE) 
		Store_field(ML_event, 1, Val_int(2));
	    else if (event.active.state & SDL_APPINPUTFOCUS)
		Store_field(ML_event, 1, Val_int(1));
	    else
		Store_field(ML_event, 1, Val_int(0));
	
	    to_return=alloc(1,0);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
	case SDL_KEYDOWN:
	case SDL_KEYUP:
	{
	    ML_event=alloc(6,0);
	    Store_field(ML_event, 0, Val_int(event.key.which));
	    Store_field(ML_event, 1, Val_int(event.key.state));
	    Store_field(ML_event, 2, Val_int(event.key.keysym.scancode));
	    Store_field(ML_event, 3, Val_int(key_to_flag[event.key.keysym.sym]));
	    Store_field(ML_event, 4, 
		mask_to_ML_flags(event.key.keysym.mod, flag_to_mod, 12));
	    Store_field(ML_event, 5, Val_int(event.key.keysym.unicode));
	    to_return=alloc(1, 1);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
	case SDL_MOUSEMOTION:
	{
	    ML_event=alloc(6,0);
	    Store_field(ML_event, 0, Val_int(event.motion.which));
	    switch(event.motion.state)
	    {
		case 0: Store_field(ML_event, 1, Val_int(0));break;
		case 1: Store_field(ML_event, 1, Val_int(1));break;
		default: Store_field(ML_event, 1, Val_int(2));break;
	    }
	    Store_field(ML_event, 2, Val_int(event.motion.x));
	    Store_field(ML_event, 3, Val_int(event.motion.y));
	    Store_field(ML_event, 4, Val_int(event.motion.xrel));
	    Store_field(ML_event, 5, Val_int(event.motion.yrel));
	    
	    to_return=alloc(1,2);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
	case SDL_MOUSEBUTTONDOWN:
	case SDL_MOUSEBUTTONUP:
	{
	    ML_event=alloc(5,0);
	    Store_field(ML_event, 0, Val_int(event.button.which));
	    switch(event.motion.state)
	    {
		case 0: Store_field(ML_event, 1, Val_int(0));break;
		case 1: Store_field(ML_event, 1, Val_int(1));break;
		default: Store_field(ML_event, 1, Val_int(2));break;
	    }
	    Store_field(ML_event, 2, Val_int(event.button.state));
	    Store_field(ML_event, 3, Val_int(event.button.x));
	    Store_field(ML_event, 4, Val_int(event.button.y));
	    
	    to_return=alloc(1,3);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
/* some joystick events must follow: 4cases */
	case SDL_VIDEORESIZE:
	{
	    ML_event=alloc(2,0);
	    Store_field(ML_event, 0, Val_int(event.resize.w));
	    Store_field(ML_event, 1, Val_int(event.resize.h));
	    
	    to_return=alloc(1,8);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
	case SDL_VIDEOEXPOSE:
	{
	    CAMLreturn(Val_int(1));
	}
	case SDL_QUIT:
	{
	    CAMLreturn(Val_int(2));
	}
	case SDL_USEREVENT:
    /* ... */
	case SDL_NUMEVENTS-1:
	{
	    ML_event=alloc(3,0);
	    Store_field(ML_event, 0, Val_int(event.user.code));
	    Store_field(ML_event, 1, (value)(event.user.data1));
	    Store_field(ML_event, 2, (value)(event.user.data2));
	    
	    to_return=alloc(1,9);
	    Store_field(to_return, 0, ML_event);
	    CAMLreturn(to_return);
	}
	case SDL_SYSWMEVENT:
	{
	    to_return=alloc(1,10);
	    Store_field(to_return, 0, (value)(event.syswm.msg));
	    CAMLreturn(to_return);
	}
	default:
	{
	    fprintf(stderr,"Unknown event.\n");
	    exit(-1);
	}
    }
}

value MLSDL_poll_event(value u)
{
    SDL_Event event;
    CAMLparam1(u);
    int isevent;
    isevent=SDL_PollEvent(&event);
    if (isevent==1) 
	CAMLreturn (SDL_event_to_ML_tevent(event));
    else CAMLreturn (Val_int(0));
}

value MLSDL_wait_event(value u)
{
    CAMLparam1(u);
    SDL_Event event;
    int isevent;
    isevent=SDL_WaitEvent(&event);
    if (isevent==1) 
	CAMLreturn (SDL_event_to_ML_tevent(event));
    else CAMLreturn (Val_int(0));
}

void MLSDL_pump_events(value u)
{
    CAMLparam1(u);
    SDL_PumpEvents();
    CAMLreturn0;
}

value MLSDL_event_state(value type, value state)
{
    CAMLparam2(type, state);
    CAMLreturn(Val_int(1+SDL_EventState(Int_val(type), Int_val(state)-1)));
}

value MLSDL_get_mouse_state(value u)
{
    CAMLparam1(u);
    CAMLlocal1(toreturn);
    int x,y,but;
    toreturn=alloc(3,0);
    but=SDL_GetMouseState(&x,&y);
    switch(but)
    {
	case 0: Store_field(toreturn, 0, Val_int(0));break;
	case 1: Store_field(toreturn, 0, Val_int(1));break;
	default: Store_field(toreturn, 0, Val_int(2));break;
    }
    
    Store_field(toreturn, 1, Val_int(x));
    Store_field(toreturn, 2, Val_int(y));
    CAMLreturn(toreturn);
}
